/*
  DB
*/

#include <stdio.h>
#include <stdlib.h>
#define _GNU_SOURCE
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h> 
#include <dirent.h>
#include <unistd.h> 
#include <ctype.h>
#include <libgen.h>
#include "subcomp.h"

/* private function */
void print_help();
void print_version(char *);
void opt_check(MAKE_IDX *, int, char **);
int cpdtodb(MAKE_IDX *, FILE *, FILE *, FILE *, FILE *, char *, int);
void standby_makeidx(MAKE_IDX *);
void clear_makeidx(MAKE_IDX *);
CPDMOL *read_CPDMOL(FILE *, int);

void print_help(){
  printf("Usage  : make_idx [option] -f <dir1> -f <dir2> ... <file> ...\n");
  printf("Options\n");
  printf("\t-a                  add to exist DB data\n");
  printf("\t-d <DBname>         specify DB name\n");
  printf("\n");
  printf("\t--help              print help\n");
  printf("\t--version           print version\n");
  printf("\n\tfile : KCF(.kcf), MDLMol(.mol), MDLSD(.sdf)\n");

  return;
}

void print_version(char *name){
  printf("%s: %s ver %s (build %s)\n", name, PACKAGE, VERSION, BUILD_ID);

  return;
}

int main(int argc, char **argv){
  int i;
  int ret = 0;
  MAKE_IDX makeidx;
  struct dirent *ent;
  struct stat st;
  char buff[BUFSIZ];
  DIR *dcpd;           /* path */
  FILE *fcpd;          /* ե */
  FILE *fidx;          /* DB(idx)ե */
  FILE *ftype;         /* DB(type)ե */
  FILE *fdat;          /* DB(dat)ե */
  int filetype;

  /* ׻ץ */
  memset(&makeidx, 0, sizeof(MAKE_IDX));

  /* β */
  opt_check(&makeidx, argc, argv);


  if (makeidx.help_mode == _HE_HELP_){
    print_help();
    clear_makeidx(&makeidx);
    exit(0);
  }
  if (makeidx.help_mode == _HE_VERSION_){
    print_version(argv[0]);
    clear_makeidx(&makeidx);
    exit(0);
  }

  /* ɬܰΥå */
  if (makeidx.cpd_fn_num + makeidx.cpd_pn_num == 0){
    printf("%s: specify compound file or directory\n", argv[0]);
    clear_makeidx(&makeidx);
    exit(0);
  }
  if (makeidx.dbname == NULL){
    printf("%s: specify DB name\n", argv[0]);
    clear_makeidx(&makeidx);
    exit(0);
  }

  /* DBե */
  sprintf(buff, "%s.idx", makeidx.dbname);
  if ((fidx = fopen(buff, (makeidx.opt_add == _YES_ ? "ab" : "wb"))) == NULL){
    printf("ERROR : cannot open DB.(%s)\n", buff);
    clear_makeidx(&makeidx);
    exit(1);
  }
  sprintf(buff, "%s.typ", makeidx.dbname);
  if ((ftype = fopen(buff, (makeidx.opt_add == _YES_ ? "ab" : "wb"))) == NULL){
    printf("ERROR : cannot open DB.(%s)\n", buff);
    clear_makeidx(&makeidx);
    fclose(fidx);
    exit(1);
  }
  sprintf(buff, "%s.dat", makeidx.dbname);
  if ((fdat = fopen(buff, (makeidx.opt_add == _YES_ ? "ab" : "wb"))) == NULL){
    printf("ERROR : cannot open DB.(%s)\n", buff);
    clear_makeidx(&makeidx);
    fclose(fidx);
    fclose(ftype);
    exit(1);
  }

  /* ǽ˲񤭹ޤʤƬ֤0LĤޤindexǶˤʤäƤޤ */
  fwrite("Nodetype", sizeof(char), 8, ftype);
  fwrite("Connect ", sizeof(char), 8, fdat);

  /* ǥ쥯ȥ롼 */
  for (i = 0; i < makeidx.cpd_pn_num; i++){
    if ((dcpd = opendir(makeidx.cpdpathname[i])) == NULL){
      perror(makeidx.cpdpathname[i]);
      clear_makeidx(&makeidx);
      fclose(fidx);
      fclose(ftype);
      fclose(fdat);
      exit(1);
    }
    while((ent = readdir(dcpd)) != NULL){
      /* ե̾γĥҥå */
      filetype = _FILE_TYPE_UNKNOWN_;
      if (strlen(ent->d_name) > 4 &&
          strncasecmp(&(ent->d_name[strlen(ent->d_name) - 4]), ".kcf", 4) == 0)
        filetype = _FILE_TYPE_KCF_;
      else
      if (strlen(ent->d_name) > 4 &&
          strncasecmp(&(ent->d_name[strlen(ent->d_name) - 4]), ".mol", 4) == 0)
        filetype = _FILE_TYPE_MDLMOL_;
      else
      if (strlen(ent->d_name) > 4 &&
          strncasecmp(&(ent->d_name[strlen(ent->d_name) - 4]), ".sdf", 4) == 0)
        filetype = _FILE_TYPE_MDLSD_;
      else
      if (strlen(ent->d_name) > 3 &&
          strncasecmp(&(ent->d_name[strlen(ent->d_name) - 3]), ".sd", 3) == 0)
        filetype = _FILE_TYPE_MDLSD_;

      if (filetype == _FILE_TYPE_UNKNOWN_) continue;

      sprintf(buff, "%s/%s", makeidx.cpdpathname[i], ent->d_name);
      stat(buff, &st);
      if (S_ISREG(st.st_mode)){
        /* оݥե */
        printf("%s\n", buff);
        if ((fcpd = fopen(buff, "r")) == NULL){
          printf("ERROR : cannot open.(%s)\n", buff);
          clear_makeidx(&makeidx);
          fclose(fidx);
          fclose(ftype);
          fclose(fdat);
          closedir(dcpd);
          exit(1);
        }
        if (cpdtodb(&makeidx, fcpd, fidx, ftype, fdat, ent->d_name, filetype) == _NO_){
          printf("ERROR : invalid file.(%s)\n", buff);
          clear_makeidx(&makeidx);
          fclose(fidx);
          fclose(ftype);
          fclose(fdat);
          closedir(dcpd);
          fclose(fcpd);
          exit(1);
        }
        fclose(fcpd);
      }
    }
    closedir(dcpd);
  }

  /* ե롼 */
  for (i = 0; i < makeidx.cpd_fn_num; i++){
    /* оݥե */
    printf("%s\n", makeidx.cpdfilename[i]);
    filetype = _FILE_TYPE_UNKNOWN_;
    if (strlen(makeidx.cpdfilename[i]) > 4 &&
        strncasecmp(&(makeidx.cpdfilename[i][strlen(makeidx.cpdfilename[i]) - 4]), ".kcf", 4) == 0)
      filetype = _FILE_TYPE_KCF_;
    else
    if (strlen(makeidx.cpdfilename[i]) > 4 &&
        strncasecmp(&(makeidx.cpdfilename[i][strlen(makeidx.cpdfilename[i]) - 4]), ".mol", 4) == 0)
      filetype = _FILE_TYPE_MDLMOL_;
    else
    if (strlen(makeidx.cpdfilename[i]) > 4 &&
        strncasecmp(&(makeidx.cpdfilename[i][strlen(makeidx.cpdfilename[i]) - 4]), ".sdf", 4) == 0)
      filetype = _FILE_TYPE_MDLSD_;
    else
    if (strlen(makeidx.cpdfilename[i]) > 3 &&
        strncasecmp(&(makeidx.cpdfilename[i][strlen(makeidx.cpdfilename[i]) - 3]), ".sd", 3) == 0)
      filetype = _FILE_TYPE_MDLSD_;

    if ((fcpd = fopen(makeidx.cpdfilename[i], "r")) == NULL){
      printf("ERROR : cannot open.(%s)\n", makeidx.cpdfilename[i]);
      clear_makeidx(&makeidx);
      fclose(fidx);
      fclose(ftype);
      fclose(fdat);
      exit(1);
    }

    if (filetype == _FILE_TYPE_UNKNOWN_){
      if (fgets(buff, sizeof(buff), fcpd) == NULL){
        printf("ERROR : invalid file.(%s)\n", makeidx.cpdfilename[i]);
        clear_makeidx(&makeidx);
        fclose(fidx);
        fclose(ftype);
        fclose(fdat);
        exit(1);
      } else {
        if (strncmp(buff, "ENTRY", 5) == 0){
          filetype = _FILE_TYPE_KCF_;
        } else {
          filetype = _FILE_TYPE_MDLSD_;
        }
        fseek(fcpd, 0L, SEEK_SET);
      }
    }

    if (cpdtodb(&makeidx, fcpd, fidx, ftype, fdat, makeidx.cpdfilename[i], filetype) == _NO_){
      printf("ERROR : invalid file.(%s)\n", buff);
      clear_makeidx(&makeidx);
      fclose(fidx);
      fclose(ftype);
      fclose(fdat);
      fclose(fcpd);
      exit(1);
    }

    fclose(fcpd);
  }
  fclose(fidx);
  fclose(ftype);
  fclose(fdat);
  clear_makeidx(&makeidx);
  exit(ret);
}

CPDMOL *read_CPDMOL(FILE *fcpd, int filetype){
  CPDMOL *cpdmol = NULL;

  switch(filetype){
  case _FILE_TYPE_KCF_:
    cpdmol = read_CPDMOL_KCF(fcpd);
    break;
  case _FILE_TYPE_MDLMOL_:
  case _FILE_TYPE_MDLSD_:
    cpdmol = read_CPDMOL_MDL(fcpd);
    break;
  }

  return cpdmol;
}

/* cpdեDB */
int cpdtodb(MAKE_IDX *makeidx, FILE *fcpd, FILE *fidx, FILE *ftype, FILE *fdat, char *pathname, int filetype){
  int i;
  long atompos;
  long bondpos;
  int order;
  int bondinfo[_BOND_ARRAY_];
  int sdf_count = 1;
  char *basec;
  char *bname;
  char idstr[_ID_LEMGTH_+ 1];
  char countstr[BUFSIZ];
  RING_LIST ringlist;

  ringlist.first = NULL;
  ringlist.last = NULL;
  ringlist.num = 0;

  while((makeidx->entry = read_CPDMOL(fcpd, filetype)) != NULL){
    if (filetype == _FILE_TYPE_KCF_){
      if (makeidx->entry->name == NULL || strlen(makeidx->entry->name) == 0){
        /* νΤ˰ꥢ */
        standby_makeidx(makeidx);
        continue;
      }
      strncpy(idstr, makeidx->entry->name, _ID_LEMGTH_);
      if (strlen(makeidx->entry->name) > _ID_LEMGTH_)
        idstr[_ID_LEMGTH_ ] = '\0';
    }

    if (filetype == _FILE_TYPE_MDLMOL_){
      basec = strdup(pathname);
      bname = basename(basec);
      if (strlen(bname) > 4 &&
          strncasecmp(&(bname[strlen(bname) - 4]), ".mol", 4) == 0)
        bname[strlen(bname) - 4] = '\0';
      strncpy(idstr, bname, _ID_LEMGTH_);
      if (strlen(bname) > _ID_LEMGTH_)
        idstr[_ID_LEMGTH_ ] = '\0';
      free(basec);
    }

    if (filetype == _FILE_TYPE_MDLSD_){
      basec = strdup(pathname);
      bname = basename(basec);
      if (strlen(bname) > 4 &&
          strncasecmp(&(bname[strlen(bname) - 4]), ".sdf", 4) == 0)
        bname[strlen(bname) - 4] = '\0';
      else
      if (strlen(bname) > 3 &&
          strncasecmp(&(bname[strlen(bname) - 3]), ".sd", 3) == 0)
        bname[strlen(bname) - 3] = '\0';
      sprintf(countstr, "-%d", sdf_count);
      strncpy(idstr, bname, _ID_LEMGTH_ - strlen(countstr));
      if (strlen(bname) > _ID_LEMGTH_ - strlen(countstr))
        idstr[_ID_LEMGTH_ - strlen(countstr)] = '\0';
      strcat(idstr, countstr);
      free(basec);
    }

    /* ӿǿ */
    calc_valence(makeidx->entry);

    switch(filetype){
    case _FILE_TYPE_KCF_:
      /* ˧ĤȽ */
      arom_checkKCF(makeidx->entry);
      break;
    case _FILE_TYPE_MDLMOL_:
    case _FILE_TYPE_MDLSD_:
      /* Ĥõ */
      searchRing(makeidx->entry, &ringlist);
      if (ringlist.num > 0){  /* Ĥ¸ߤʤ */
        /* ˧Ĥõ */
        arom_checkMOL(makeidx->entry, &ringlist);
      }
      /* ̤atombond˳ǼΤǡĤΥꥹȤϤ⤦ʤ */
      free_RINGLIST(&ringlist);
      break;
    }

    /* ƸҤǷֹ򥽡 */
    sort_bondatom(makeidx->entry);

    /* ƤȽ() */
    checkNegaPos(makeidx->entry, makeidx->opt_dim);

    /* ƤȽ(ŷ) */
    checkCisTrans(makeidx->entry);

/*
for (i = 0; i < makeidx->entry->atomnum; i++){
int j;
printf("%2d %-3s %-3s %2d %2d %d %2d %d %d %d %d\n", i + 1,
makeidx->entry->atom[i].nid,
makeidx->entry->atom[i].aid,
makeidx->entry->atom[i].atomno,
makeidx->entry->atom[i].charge,
makeidx->entry->atom[i].valence,
makeidx->entry->atom[i].chiral,
makeidx->entry->atom[i].numDefi,
makeidx->entry->atom[i].numH,
makeidx->entry->atom[i].arom,
makeidx->entry->atom[i].bondnum);
for (j = 0; j < makeidx->entry->atom[i].bondnum; j++){
printf(" %2d", makeidx->entry->atom[i].bond[j] + 1);
}
printf("\n");
}

for (i = 0; i < makeidx->entry->bondnum; i++){
printf("%2d %2d %2d %d %d %d %d\n", i + 1,
makeidx->entry->bond[i].a1,
makeidx->entry->bond[i].a2,
makeidx->entry->bond[i].order,
makeidx->entry->bond[i].arom,
makeidx->entry->bond[i].stereo,
makeidx->entry->bond[i].geometry);
}
*/
    /* atomsbondsγ */
    makeidx->atoms = mallocArray_char2(makeidx->entry->atomnum, _ATOM_ARRAY_);
    if (makeidx->atoms == NULL){
      return _NO_;
    }

    /* type */
    for (i = 0 ; i < makeidx->entry->atomnum; i++){
      makeidx->atoms[i][_ATOM_ATOMNO_]   = (char)makeidx->entry->atom[i].atomno;
      makeidx->atoms[i][_ATOM_HYDROGEN_] = (char)(makeidx->entry->atom[i].numH + makeidx->entry->atom[i].numDefi);
      makeidx->atoms[i][_ATOM_CHARGE_]   = (char)(makeidx->entry->atom[i].charge + _CHARGE_OFFSET_);
      makeidx->atoms[i][_ATOM_VALENCE_]  = (char)makeidx->entry->atom[i].valence;
      makeidx->atoms[i][_ATOM_CHIRAL_]   = (char)makeidx->entry->atom[i].chiral;
    }

    /* atomsν񤭹 */
    /*
     int                           ҿ
     int                           ҿ(ǰʳ)
     int                           
     char * ҿ * _TYPE_ARRAY_  Ҥξ
    */
    atompos = ftell(ftype);      /* Ƭ */

    fwrite(&(makeidx->entry->atomnum), sizeof(int), 1, ftype);
    fwrite(&(makeidx->entry->atomnum_sub), sizeof(int), 1, ftype);
    fwrite(&(makeidx->entry->bondnum), sizeof(int), 1, ftype);
    for (i = 0 ; i < makeidx->entry->atomnum; i++){
      fwrite(&(makeidx->atoms[i][0]), sizeof(char), _ATOM_ARRAY_, ftype);
    }
    fflush(ftype);

    /* bondsν񤭹 */
    /*
     int *  * _BOND_ARRAY_  ξ
    */
    bondpos = ftell(fdat);    /* Ƭ */

    for (i = 0; i < makeidx->entry->bondnum; i++){
      bondinfo[_BOND_ATOM1_] = makeidx->entry->bond[i].a1;
      bondinfo[_BOND_ATOM2_] = makeidx->entry->bond[i].a2;
      if (makeidx->entry->bond[i].arom == _YES_){
        order = 4;
      } else {
        order = makeidx->entry->bond[i].order;
      }

      switch(order){
      case 1:
        bondinfo[_BOND_ORDER_] = 0x1;
        break;
      case 2:
        bondinfo[_BOND_ORDER_] = 0x2;
        break;
      case 3:
        bondinfo[_BOND_ORDER_] = 0x4;
        break;
      case 4:
        bondinfo[_BOND_ORDER_] = 0x8;
        break;
      default:
        bondinfo[_BOND_ORDER_] = 0x1;
        break;
      }
      bondinfo[_BOND_GEOMETRY_] = makeidx->entry->bond[i].geometry;
      fwrite(bondinfo, sizeof(int), _BOND_ARRAY_, fdat);
    }
    fflush(fdat);

    /* indexν񤭹 */
    /*
       char[_ID_LEMGTH_]             KEGG_ID or MDLե̾
       long                          ҾΥեå
       long                          Υեå
    */
    fwrite(idstr, sizeof(char), _ID_LEMGTH_, fidx);
    fwrite(&(atompos), sizeof(long), 1, fidx);
    fwrite(&(bondpos), sizeof(long), 1, fidx);

    /* νΤ˰ꥢ */
    standby_makeidx(makeidx);
    sdf_count++;

    /* MDLMolեʤ1󤷤ɤޤʤ */
    if (filetype == _FILE_TYPE_MDLMOL_) break;
  }

  return _YES_;
}

/*
getopt_long()λѤͤ
ɬפgetopt.h¿ʬɡ
ϥեμͳ٤ʤʤ뤿ᡢ
缡ˤƹԤ
*/
void opt_check(MAKE_IDX *makeidx, int argc, char** argv){
  int i;

  makeidx->cpdfilename = (char **)malloc(sizeof(char *) * (argc - 1));
  makeidx->cpdpathname = (char **)malloc(sizeof(char *) * (argc - 1));
  if (makeidx->cpdfilename == NULL || makeidx->cpdpathname == NULL) return;
  memset(makeidx->cpdfilename, 0, sizeof(char *) * (argc - 1));
  memset(makeidx->cpdpathname, 0, sizeof(char *) * (argc - 1));

  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-f") == 0) {
      if (i + 1 < argc) {
        StrAllocCopy(&makeidx->cpdpathname[makeidx->cpd_pn_num], argv[i + 1]);
        makeidx->cpd_pn_num++;
        i++;
      }
    } else
    if (strcmp(argv[i], "-d") == 0) {
      if (i + 1 < argc) {
        StrAllocCopy(&makeidx->dbname, argv[i + 1]);
        i++;
      }
    } else
    if (strcmp(argv[i], "-a") == 0) {
      makeidx->opt_add = _YES_;
    } else
    if (strcmp(argv[i], "--help") == 0) {
      makeidx->help_mode = _HE_HELP_;
    } else
    if (strcmp(argv[i], "--version") == 0) {
      makeidx->help_mode = _HE_VERSION_;
    } else
    if (strncmp(argv[i], "-", 1) == 0) {
      printf("ERROR : Illegal option \"%s\"\n", argv[i]);
      break;
    } else {
      StrAllocCopy(&makeidx->cpdfilename[makeidx->cpd_fn_num], argv[i]);
      makeidx->cpd_fn_num++;
    }
  }

  return;
}

/* μ¹ԤΤ˰ꥢ */
void standby_makeidx(MAKE_IDX *makeidx){
  if (makeidx->atoms != NULL){
    freeArray_char2(&(makeidx->atoms), makeidx->entry->atomnum);
  }

  if (makeidx->entry != NULL){
    clear_CPDMOL(makeidx->entry);
    free(makeidx->entry);
    makeidx->entry = NULL;
  }

  return;
}

/* λθդ */
void clear_makeidx(MAKE_IDX *makeidx){
  int i;

  for (i = 0; i < makeidx->cpd_fn_num; i++){
    if (makeidx->cpdfilename[i] != NULL) free(makeidx->cpdfilename[i]);
  }
  if (makeidx->cpdfilename != NULL) free(makeidx->cpdfilename);
  for (i = 0; i < makeidx->cpd_pn_num; i++){
    if (makeidx->cpdpathname[i] != NULL) free(makeidx->cpdpathname[i]);
  }
  if (makeidx->cpdpathname != NULL) free(makeidx->cpdpathname);

  if (makeidx->dbname != NULL) free(makeidx->dbname);

  if (makeidx->atoms != NULL){
    freeArray_char2(&(makeidx->atoms), makeidx->entry->atomnum);
  }

  if (makeidx->entry != NULL){
    clear_CPDMOL(makeidx->entry);
    free(makeidx->entry);
    makeidx->entry = NULL;
  }

  return;
}
