/*
  DB
*/

#include <stdio.h>
#include <stdlib.h>
#define _GNU_SOURCE
#include <string.h>
#include <ctype.h>
#include <libgen.h>
#include "subcomp.h"

/* private function */
void print_help();
void print_version(char *);
int opt_check(SUBCOMP *, int, char **);
int exec_subcomp(SUBCOMP *, int);
void output_result(SUBCOMP *);
void standby_subcomp(SUBCOMP *, int);
void clear_subcomp(SUBCOMP *);
void sortRESULT(RESULT **, int);
void printRESULT(SUBCOMP *, FILE *);
void freeRESULTlist(RESULT **);

static FILE *fcpd = NULL;
static FILE *fout = NULL;
static FILE *fidx = NULL;
static FILE *fatom = NULL;
static FILE *fbond = NULL;

void print_help(){
  printf("Summary : 'SUB'structure search program.\n");
  printf("Usage  : subcomp [option] -q <file> -d <DBname>\n");
  printf("Options\n");
  printf("\t-d <DBname>         specify DB name\n");
  printf("\t-q <file>           query data\n");
  printf("\t                    KCF(.kcf), MDLMol(.mol), MDLSD(.sdf)\n");
  printf("\n");
  printf("matchingspecifier:\n");
  printf("\t-m/--match <mode>   matching mode(all)\n");
  printf("\t                      all (structure search mode)\n");
  printf("\t                      sub (SUBstructure search mode)\n");
  printf("\t                      sup (SUPERstructure search mode)\n");
  printf("\n");
  printf("\t--charge            consider charged atom (ignore)\n");
  printf("\t--coordinate        consider coordinate bond (ignore)\n");
  printf("\t--valence           consider atom valence (ignore)\n");
  printf("\t--chiral            consider chiral (ignore)\n");
  printf("\t--kcfoutput         output in the KCF RPAIR format\n");
  printf("\t-o <file>           output all computational result (stdout)\n");
  printf("\n");
  printf("\t--help              print help\n");
  printf("\t--version           print version\n");

  return;
}

void print_version(char *name){
  printf("%s: %s ver %s (build %s)\n", name, PACKAGE, VERSION, BUILD_ID);

  return;
}

int main(int argc, char **argv){
  int ret = 0;
  SUBCOMP subcomp;
  char buff[BUFSIZ];
  int filetype;
  char *basec;
  char *bname;

  fout = stdout;

  /* ׻ץ */
  memset(&subcomp, 0, sizeof(SUBCOMP));

  subcomp.match_mode = _MA_ALL_;

  /* β */
  if (opt_check(&subcomp, argc, argv) == _NO_){
    clear_subcomp(&subcomp);
    exit(1);
  }

  if (subcomp.help_mode == _HE_HELP_){
    print_help();
    clear_subcomp(&subcomp);
    exit(0);
  }
  if (subcomp.help_mode == _HE_VERSION_){
    print_version(argv[0]);
    clear_subcomp(&subcomp);
    exit(0);
  }

  /* ɬܰΥå */
  if (subcomp.qfilename == NULL){
    printf("%s: specify QUERY file\n", argv[0]);
    clear_subcomp(&subcomp);
    exit(1);
  }
  if (subcomp.dbname == NULL){
    printf("%s: specify DB name\n", argv[0]);
    clear_subcomp(&subcomp);
    exit(1);
  }

  /* Υץ */
  if (subcomp.output_fn != NULL){
    if ((fout = fopen(subcomp.output_fn, "w")) == NULL){
      exit(1);
    }
  }

  /* DBե */
  sprintf(buff, "%s.idx", subcomp.dbname);
  if ((fidx = fopen(buff, "rb")) == NULL){
    printf("ERROR : cannot open DB.(%s)\n", buff);
    clear_subcomp(&subcomp);
    exit(1);
  }
  sprintf(buff, "%s.typ", subcomp.dbname);
  if ((fatom = fopen(buff, "rb")) == NULL){
    printf("ERROR : cannot open DB.(%s)\n", buff);
    clear_subcomp(&subcomp);
    fclose(fidx);
    exit(1);
  }
  sprintf(buff, "%s.dat", subcomp.dbname);
  if ((fbond = fopen(buff, "rb")) == NULL){
    printf("ERROR : cannot open DB.(%s)\n", buff);
    clear_subcomp(&subcomp);
    fclose(fidx);
    fclose(fatom);
    exit(1);
  }

  /* ꡼ե */
  if ((fcpd = fopen(subcomp.qfilename, "r")) == NULL){
    printf("ERROR : cannot open.(%s)\n", subcomp.qfilename);
    perror(subcomp.qfilename);
    clear_subcomp(&subcomp);
    exit(1);
  }

  /* ե̾γĥҥå */
  filetype = _FILE_TYPE_UNKNOWN_;
  if (strlen(subcomp.qfilename) > 4 &&
      strncasecmp(&(subcomp.qfilename[strlen(subcomp.qfilename) - 4]), ".kcf", 4) == 0)
    filetype = _FILE_TYPE_KCF_;
  else
  if (strlen(subcomp.qfilename) > 4 &&
      strncasecmp(&(subcomp.qfilename[strlen(subcomp.qfilename) - 4]), ".mol", 4) == 0)
    filetype = _FILE_TYPE_MDLMOL_;
  else
  if (strlen(subcomp.qfilename) > 4 &&
      strncasecmp(&(subcomp.qfilename[strlen(subcomp.qfilename) - 4]), ".sdf", 4) == 0)
    filetype = _FILE_TYPE_MDLSD_;
  else
  if (strlen(subcomp.qfilename) > 3 &&
      strncasecmp(&(subcomp.qfilename[strlen(subcomp.qfilename) - 3]), ".sd", 3) == 0)
    filetype = _FILE_TYPE_MDLSD_;

  /* ĥҤǤ狼ʤΤ1ܤå */
  if (filetype == _FILE_TYPE_UNKNOWN_){
    if (fgets(buff, sizeof(buff), fcpd) == NULL){
      printf("ERROR : invalid file.(%s)\n", subcomp.qfilename);
      fclose(fcpd);
      fclose(fidx);
      fclose(fatom);
      fclose(fbond);
      clear_subcomp(&subcomp);
      fclose(fout);
      exit(1);
    } else {
      if (strncmp(buff, "ENTRY", 5) == 0){
        filetype = _FILE_TYPE_KCF_;
      } else {
        filetype = _FILE_TYPE_MDLSD_;
      }
      fseek(fcpd, 0L, SEEK_SET);
    }
  }
  /* ꡼ɤ߹ */
  switch(filetype){
  case _FILE_TYPE_KCF_:
    subcomp.query = read_CPDMOL_KCF(fcpd);
    break;
  case _FILE_TYPE_MDLMOL_:
  case _FILE_TYPE_MDLSD_:
    subcomp.query = read_CPDMOL_MDL(fcpd);
    subcomp.query->name = (char *)malloc(sizeof(char) * (_ID_LEMGTH_ + 1));
    basec = strdup(subcomp.qfilename);
    bname = basename(basec);
    if (strlen(bname) > 4 &&
        strncasecmp(&(bname[strlen(bname) - 4]), ".mol", 4) == 0)
      bname[strlen(bname) - 4] = '\0';
    else
    if (strlen(bname) > 4 &&
        strncasecmp(&(bname[strlen(bname) - 4]), ".sdf", 4) == 0)
      bname[strlen(bname) - 4] = '\0';
    else
    if (strlen(bname) > 3 &&
        strncasecmp(&(bname[strlen(bname) - 3]), ".sd", 3) == 0)
      bname[strlen(bname) - 3] = '\0';
    strncpy(subcomp.query->name, bname, _ID_LEMGTH_);
    if (strlen(bname) > _ID_LEMGTH_)
      subcomp.query->name[_ID_LEMGTH_ ] = '\0';
    free(basec);
    break;
  }

  if (subcomp.query->atomnum == 0){
    printf("%s: can not read KCF file\n", subcomp.qfilename);
    fclose(fcpd);
    fclose(fidx);
    fclose(fatom);
    fclose(fbond);
    clear_subcomp(&subcomp);
    fclose(fout);
    exit(1);
  }

  /* ʬ¤ */
  ret = exec_subcomp(&subcomp, filetype);

  if (ret == 0) output_result(&subcomp);

  clear_subcomp(&subcomp);
  fclose(fcpd);
  fclose(fidx);
  fclose(fatom);
  fclose(fbond);
  fclose(fout);
  exit(ret);
}

int exec_subcomp(SUBCOMP* subcomp, int filetype){
  int i;
  int ret = 0;
  int mode = 0;
  int atomnum = 0;
  int atomnum_sub = 0;
  int bondnum = 0;
  int order = 0;
  RING_LIST ringlist;

  ringlist.first = NULL;
  ringlist.last = NULL;
  ringlist.num = 0;

  long pos[2];

  pos[0] = pos[1] = 0L;

  /* ӿǿ */
  calc_valence(subcomp->query);

  switch(filetype){
  case _FILE_TYPE_KCF_:
    /* ˧ĤȽ */
    arom_checkKCF(subcomp->query);
    break;
  case _FILE_TYPE_MDLMOL_:
  case _FILE_TYPE_MDLSD_:
    /* Ĥõ */
    searchRing(subcomp->query, &ringlist);
    if (ringlist.num > 0){  /* Ĥ¸ߤʤ */
      /* ˧Ĥõ */
      arom_checkMOL(subcomp->query, &ringlist);
    }
    /* ̤atombond˳ǼΤǡĤΥꥹȤϤ⤦ʤ */
    free_RINGLIST(&ringlist);
    break;
  }

  /* ƸҤǷֹ򥽡 */
  sort_bondatom(subcomp->query);

  /* ƤȽ() */
  checkNegaPos(subcomp->query, 0);

  /* ƤȽ(ŷ) */
  checkCisTrans(subcomp->query);

  /* atomsbondmatrixγ */
  subcomp->Qatoms = mallocArray_char2(subcomp->query->atomnum, _ATOM_ARRAY_);
  subcomp->Qbonds = mallocArray_int2(subcomp->query->bondnum, _BOND_ARRAY_);
  subcomp->Qbondmatrix = mallocArray_char2(subcomp->query->atomnum, subcomp->query->atomnum);
  if (subcomp->Qatoms == NULL || subcomp->Qbondmatrix == NULL){
    printf("can not allocate query data\n");
    return -1;
  }

  /* atom */
  for (i = 0; i < subcomp->query->atomnum; i++){
    subcomp->Qatoms[i][_ATOM_ATOMNO_]   = (char)subcomp->query->atom[i].atomno;
    subcomp->Qatoms[i][_ATOM_CHARGE_]   = (char)(subcomp->query->atom[i].charge + _CHARGE_OFFSET_);
    subcomp->Qatoms[i][_ATOM_VALENCE_]  = (char)subcomp->query->atom[i].valence;
    subcomp->Qatoms[i][_ATOM_CHIRAL_]   = (char)subcomp->query->atom[i].chiral;
    /* _ATOM_HYDROGEN_ϸꤹ(SUPSUBͤ㤦) */
  }

  /* bond */
  for (i = 0; i < subcomp->query->bondnum; i++){
    subcomp->Qbonds[i][_BOND_ATOM1_]     = subcomp->query->bond[i].a1;
    subcomp->Qbonds[i][_BOND_ATOM2_]     = subcomp->query->bond[i].a2;
    if (subcomp->query->bond[i].arom == _YES_){
      order = 4;
    } else {
      order = subcomp->query->bond[i].order;
    }
    switch(order){
    case 1:
      subcomp->Qbonds[i][_BOND_ORDER_]   = 0x1;
      break;
    case 2:
      subcomp->Qbonds[i][_BOND_ORDER_]   = 0x2;
      break;
    case 3:
      subcomp->Qbonds[i][_BOND_ORDER_]   = 0x4;
      break;
    case 4:
      subcomp->Qbonds[i][_BOND_ORDER_]   = 0x8;
      break;
    default:
      subcomp->Qbonds[i][_BOND_ORDER_]   = 0x1;
      break;
    }
    subcomp->Qbonds[i][_BOND_GEOMETRY_]  = subcomp->query->bond[i].geometry;
  }

  /* ۰̷ */
  if (subcomp->opt_coordinate == _YES_)
    correctCoordinate(subcomp->query->bondnum, subcomp->Qatoms, subcomp->Qbonds);

  /* bondmatrix */
  make_bondmatrix(subcomp->query->bondnum, subcomp->Qbonds, subcomp->Qbondmatrix);

  subcomp->m_row = subcomp->query->atomnum;

/*
printf("atom %d : bond %d\n", subcomp->query->atomnum, subcomp->query->atomnum);
printf("[atom]\n");
for (i = 0; i < subcomp->query->atomnum; i++){
printf(" %d, %d, %d, %d, %d\n", subcomp->Qatoms[i][0], subcomp->Qatoms[i][1], subcomp->Qatoms[i][2], subcomp->Qatoms[i][3], subcomp->Qatoms[i][4]);
}
printf("[bond]\n");
for (i = 0; i < subcomp->query->bondnum; i++){
printf(" %d, %d, %d, %d\n", subcomp->Qbonds[i][0], subcomp->Qbonds[i][1], subcomp->Qbonds[i][2], subcomp->Qbonds[i][3]);
}
*/

  /* indexꥨȥ꡼Ф */
  for(fread(subcomp->idstr, sizeof(char), _ID_LEMGTH_, fidx),
      fread(pos, sizeof(long), 2, fidx),
      subcomp->idstr[_ID_LEMGTH_] = '\0';
        feof(fidx) == 0;
          standby_subcomp(subcomp, mode),
          fread(subcomp->idstr, sizeof(char), _ID_LEMGTH_, fidx),
          fread(pos, sizeof(long), 2, fidx),
          subcomp->idstr[_ID_LEMGTH_] = '\0'){

    if (pos[0] == 0L || pos[1] == 0L) continue;

//printf("[ID] : %s (%ld, %ld)\n", subcomp->idstr, pos[0], pos[1]);

    /* DBȥ꡼θҿ */
    fseek(fatom, pos[0], SEEK_SET);
    fread(&atomnum, sizeof(int), 1, fatom);
    fread(&atomnum_sub, sizeof(int), 1, fatom);
    fread(&bondnum, sizeof(int), 1, fatom);

    /* ǸҤΥȥ꡼ϥå */
    if (atomnum_sub == 0) continue;

    /* ҿå(ǤΤȤηƥȤƤʤΤǷϸƨ) */
    if (subcomp->query->atomnum_sub <= atomnum_sub && subcomp->match_mode != _MA_SUP_){
      /* SUBstructure⡼ */
      mode = _MA_SUB_;
      subcomp->Patomnum     = subcomp->query->atomnum;      /* ʬ¤θҿ */
      subcomp->Patomnum_sub = subcomp->query->atomnum_sub;  /* ʬ¤θҿ(ǽ) */
      subcomp->Pbondnum     = subcomp->query->bondnum;      /* ʬ¤η */
      subcomp->Watomnum     = atomnum;          /* ι¤θҿ */
      subcomp->Watomnum_sub = atomnum_sub;      /* ι¤θҿ(ǽ) */
      subcomp->Wbondnum     = bondnum;          /* ι¤η */
      subcomp->m_row = subcomp->Patomnum;
      subcomp->m_col = subcomp->Watomnum;  /* ޥȥå  bitĹҿ(WatomnumƱ) */
      subcomp->m_length = (subcomp->m_col + _BIT_LENGTH_ - 1) / _BIT_LENGTH_;

      subcomp->Patoms       = subcomp->Qatoms;
      subcomp->Pbonds       = subcomp->Qbonds;
      subcomp->Pbondmatrix  = subcomp->Qbondmatrix;

      /* typeκ */
      for (i = 0; i < subcomp->query->atomnum; i++){
        subcomp->Qatoms[i][_ATOM_HYDROGEN_] = (char)subcomp->query->atom[i].numH;
      }

      /* DBȥ꡼atoms */
      subcomp->Watoms = mallocArray_char2(subcomp->Watomnum, _ATOM_ARRAY_);
      if (subcomp->Watoms == NULL){
        ret = -1;
        break;
      }
      for (i = 0; i < subcomp->Watomnum; i++){
        fread(subcomp->Watoms[i], sizeof(char), _ATOM_ARRAY_, fatom);
      }

      /* DBȥ꡼bonds */
      if (subcomp->Wbondnum > 0){
        subcomp->Wbonds = mallocArray_int2(subcomp->Wbondnum, _BOND_ARRAY_);
        if (subcomp->Wbonds == NULL){
          ret = -1;
          break;
        }
        fseek(fbond, pos[1], SEEK_SET);
        for (i = 0; i < subcomp->Wbondnum; i++){
          fread(subcomp->Wbonds[i], sizeof(int), _BOND_ARRAY_, fbond);
        }
      }

      /* DBȥ꡼bondmatrix */
      subcomp->Wbondmatrix = mallocArray_char2(subcomp->Watomnum, subcomp->Watomnum);
      if (subcomp->Wbondmatrix == NULL){
        ret = -1;
        break;
      }

      /* ۰̷ */
      if (subcomp->opt_coordinate == _YES_)
        correctCoordinate(subcomp->Wbondnum, subcomp->Watoms, subcomp->Wbonds);

      make_bondmatrix(subcomp->Wbondnum, subcomp->Wbonds, subcomp->Wbondmatrix);
    } else
    if (subcomp->query->atomnum_sub >= atomnum_sub && subcomp->match_mode != _MA_SUB_){
      /* SUPERstructure⡼ */
      mode = _MA_SUP_;
      subcomp->Patomnum     = atomnum;          /* ʬ¤θҿ */
      subcomp->Patomnum_sub = atomnum_sub;      /* ʬ¤θҿ(ǽ) */
      subcomp->Pbondnum     = bondnum;          /* ʬ¤η */
      subcomp->Watomnum     = subcomp->query->atomnum;      /* ι¤θҿ */
      subcomp->Watomnum_sub = subcomp->query->atomnum_sub;  /* ι¤θҿ(ǽ) */
      subcomp->Wbondnum     = subcomp->query->bondnum;      /* ι¤η */
      subcomp->m_row = subcomp->Patomnum;
      subcomp->m_col = subcomp->Watomnum;  /* ޥȥå  bitĹҿ(WatomnumƱ) */
      subcomp->m_length = (subcomp->m_col + _BIT_LENGTH_ - 1) / _BIT_LENGTH_;

      subcomp->Watoms       = subcomp->Qatoms;
      subcomp->Wbonds       = subcomp->Qbonds;
      subcomp->Wbondmatrix  = subcomp->Qbondmatrix;

      /* typeκ */
      for (i = 0; i < subcomp->query->atomnum; i++){
        subcomp->Qatoms[i][_ATOM_HYDROGEN_] = (char)(subcomp->query->atom[i].numH + subcomp->query->atom[i].numDefi);
      }

      /* DBȥ꡼atoms */
      subcomp->Patoms = mallocArray_char2(subcomp->Patomnum, _ATOM_ARRAY_);
      if (subcomp->Patoms == NULL){
        ret = -1;
        break;
      }
      for (i = 0; i < subcomp->Patomnum; i++){
        fread(subcomp->Patoms[i], sizeof(char), _ATOM_ARRAY_, fatom);
      }

      /* DBȥ꡼bonds */
      if (subcomp->Pbondnum > 0){
        subcomp->Pbonds = mallocArray_int2(subcomp->Pbondnum, _BOND_ARRAY_);
        if (subcomp->Pbonds == NULL){
          ret = -1;
          break;
        }
        fseek(fbond, pos[1], SEEK_SET);
        for (i = 0; i < subcomp->Pbondnum; i++){
          fread(subcomp->Pbonds[i], sizeof(int), _BOND_ARRAY_, fbond);
        }
      }

      /* DBȥ꡼bondmatrix */
      subcomp->Pbondmatrix = mallocArray_char2(subcomp->Patomnum, subcomp->Patomnum);
      if (subcomp->Pbondmatrix == NULL){
        ret = -1;
        break;
      }

      /* ۰̷ */
      if (subcomp->opt_coordinate == _YES_)
        correctCoordinate(subcomp->Pbondnum, subcomp->Patoms, subcomp->Pbonds);

      make_bondmatrix(subcomp->Pbondnum, subcomp->Pbonds, subcomp->Pbondmatrix);
    } else {
      continue;
    }
    /* ޥå󥰥ޥȥåƹ */
    if (resetMATRIX(&(subcomp->m), subcomp->m_row, subcomp->m_col) == _NO_) break;

    setMATRIX(subcomp, mode);

    if (resetMATRIXARRAY(&(subcomp->mm), subcomp->m_row, subcomp->m_row, subcomp->m_col) == _NO_) break;

    /* targetλѥե饰 */
    subcomp->f_used = (unsigned int *)malloc(sizeof(unsigned int) * subcomp->m_length);
    if (subcomp->f_used == NULL) break;
    memset(subcomp->f_used, 0, sizeof(unsigned int) * subcomp->m_length);

    /* ޥå */
    search(subcomp);
  }
  return ret;
}

void output_result(SUBCOMP* subcomp){
  RESULT *result;
  int count;

  if (subcomp->resultnum > 0){
    /* ̥ꥹȤ */
    subcomp->results = (RESULT **)malloc(sizeof(RESULT *) * (subcomp->resultnum));
    if (subcomp->results != NULL){
      memset(subcomp->results, 0, sizeof(RESULT *) * (subcomp->resultnum));

      count = 0;
      result = subcomp->result_top;
      while(result != NULL){
        subcomp->results[count] = result;

        count++;
        result = result->next;
      }

      /* ̤Υ */
      sortRESULT(subcomp->results, subcomp->resultnum);

      /* ̽ */
      printRESULT(subcomp, fout);
    }
  }
}

/*
getopt_long()λѤͤ
ɬפgetopt.h¿ʬɡ
ϥեμͳ٤ʤʤ뤿ᡢ
缡ˤƹԤ
*/
int opt_check(SUBCOMP *subcomp, int argc, char** argv){
  int i;
  int ret = _YES_;

  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-o") == 0 ||
        strcmp(argv[i], "--output") == 0) {
      if (i + 1 < argc) {
        StrAllocCopy(&subcomp->output_fn, argv[i + 1]);
        i++;
      }
    } else
    if (strcmp(argv[i], "-q") == 0) {
      if (i + 1 < argc) {
        StrAllocCopy(&subcomp->qfilename, argv[i + 1]);
        i++;
      }
    } else
    if (strcmp(argv[i], "-d") == 0) {
      if (i + 1 < argc) {
        StrAllocCopy(&subcomp->dbname, argv[i + 1]);
        i++;
      }
    } else
    /* match specifier */
    if (strcmp(argv[i], "-m") == 0 ||
        strcmp(argv[i], "--match") == 0) {
      if (i + 1 < argc) {
        if (strcasecmp(argv[i + 1], "sub") == 0)
          subcomp->match_mode = _MA_SUB_;
        else
        if (strcasecmp(argv[i + 1], "sup") == 0)
          subcomp->match_mode = _MA_SUP_;
        else
        if (strcasecmp(argv[i + 1], "all") == 0)
          subcomp->match_mode = _MA_ALL_;
        i++;
      }
    } else
    if (strcmp(argv[i], "--charge") == 0) {
      subcomp->opt_charge = _YES_;
    } else
    if (strcmp(argv[i], "--coordinate") == 0) {
      subcomp->opt_coordinate = _YES_;
    } else
    if (strcmp(argv[i], "--valence") == 0) {
      subcomp->opt_valence = _YES_;
    } else
    if (strcmp(argv[i], "--chiral") == 0) {
      subcomp->opt_chiral = _CC_BOTH_;
    } else
    if (strcmp(argv[i], "--kcfoutput") == 0) {
      subcomp->opt_kcfoutput = _YES_;
    } else
    if (strcmp(argv[i], "--help") == 0) {
      subcomp->help_mode = _HE_HELP_;
    } else
    if (strcmp(argv[i], "--version") == 0) {
      subcomp->help_mode = _HE_VERSION_;
    } else
    if (strncmp(argv[i], "-", 1) == 0) {
      printf("ERROR : Illegal option \"%s\"\n", argv[i]);
      ret = _NO_;
      break;
    }
  }

  return ret;
}

/* SUBCOMPμ¹ԤΤ˰ꥢ */
void standby_subcomp(SUBCOMP *subcomp, int mode){

  switch(mode){
  case _MA_SUB_:
    freeArray_char2(&(subcomp->Watoms), subcomp->Watomnum);
    freeArray_int2(&(subcomp->Wbonds), subcomp->Wbondnum);
    freeArray_char2(&(subcomp->Wbondmatrix), subcomp->Watomnum);
    break;
  case _MA_SUP_:
    freeArray_char2(&(subcomp->Patoms), subcomp->Patomnum);
    freeArray_int2(&(subcomp->Pbonds), subcomp->Pbondnum);
    freeArray_char2(&(subcomp->Pbondmatrix), subcomp->Patomnum);
    break;
  }
  subcomp->Patoms = NULL;
  subcomp->Pbonds = NULL;
  subcomp->Pbondmatrix = NULL;
  subcomp->Watoms = NULL;
  subcomp->Wbonds = NULL;
  subcomp->Wbondmatrix = NULL;

  if (subcomp->m != NULL){
    freeArray_uint2(&(subcomp->m), subcomp->m_row);
  }
  if (subcomp->mm != NULL) freeArray_uint3(&(subcomp->mm), subcomp->m_row, subcomp->m_row);

  if (subcomp->f_used != NULL){
    free(subcomp->f_used);
    subcomp->f_used = NULL;
  }
  subcomp->m_length = 0;

  return;
}

/* λθդ */
void clear_subcomp(SUBCOMP *subcomp){

  if (subcomp->qfilename != NULL){
    free(subcomp->qfilename);
  }
  if (subcomp->output_fn != NULL){
    free(subcomp->output_fn);
  }
  if (subcomp->dbname != NULL){
    free(subcomp->dbname);
  }

/*
  freeArray_char2(&(subcomp->Patoms), subcomp->Patomnum);
  freeArray_int2(&(subcomp->Pbonds), subcomp->Pbondnum);
  freeArray_char2(&(subcomp->Pbondmatrix), subcomp->Patomnum);
  freeArray_char2(&(subcomp->Watoms), subcomp->Watomnum);
  freeArray_int2(&(subcomp->Wbonds), subcomp->Wbondnum);
  freeArray_char2(&(subcomp->Wbondmatrix), subcomp->Watomnum);
*/

  if (subcomp->query != NULL){
    clear_CPDMOL(subcomp->query);
    free(subcomp->query);
  }

  if (subcomp->m != NULL) freeArray_uint2(&(subcomp->m), subcomp->m_row);
  if (subcomp->mm != NULL) freeArray_uint3(&(subcomp->mm), subcomp->m_row, subcomp->m_row);
  if (subcomp->f_used != NULL){
    free(subcomp->f_used);
  }

  if (subcomp->result_top != NULL) freeRESULTlist(&(subcomp->result_top));
  if (subcomp->results != NULL){
    free(subcomp->results);
  }

  return;
}

/* scoreǹ߽(shell sort) */
/* ǡϤʤQuick Sortɬʥ */
void sortRESULT(RESULT **results, int num){
  int i, j, k;
  int h;
  RESULT *r;

  i = j = k = 0;
  h = 1;
  while(h <= num) h = h * 3 + 1;
  h = h / 9;
  if (h < 1) h = 1;

  while(h > 0){
    for (i = h; i < num; i++){
      k = i;
      j = i - h;
      while(j >= 0 && 
           (results[j]->score < results[k]->score ||
            (results[j]->score == results[k]->score &&
             strcmp(results[j]->idstr, results[k]->idstr) > 0))){
        r          = results[j];
        results[j] = results[k];
        results[k] = r;
        j -= h;
        k -= h;
      }
    }
    h = h / 3;
  }
}

void printRESULT(SUBCOMP *subcomp, FILE *fout){
  int i, j;
  int count;
  RESULT *result;
  int pairNO;
  int atomNO;
  char buff[BUFSIZ];

  for (i = 0; i < subcomp->resultnum; i++){
    result = subcomp->results[i];

    if (subcomp->opt_kcfoutput == _YES_){
      fprintf(fout, "ENTRY       RP00000           ReactantPair\n");
      fprintf(fout, "NAME        %s_%s\n", subcomp->query->name, result->idstr);
      fprintf(fout, "COMPOUND    %s %s\n", subcomp->query->name, result->idstr);
      count = 1;

      switch(result->mode){
      case _MA_SUB_:
        fprintf(fout, "ALIGN       %d  #similarity=%f\n", subcomp->query->atomnum_sub, result->score);
        for (j = 0; j < subcomp->query->atomnum; j++){
          pairNO = result->map[j];
          atomNO = result->atomicNO[j];

          if (pairNO < 0) continue;
          getLabel(atomNO, buff);
          fprintf(fout, "            %-4d%3d:%-3s %3d:%-3s\n", count, j + 1, buff, pairNO + 1, buff);
          count++;
        }
        break;
      case _MA_SUP_:
        fprintf(fout, "ALIGN       %d  #similarity=%f\n", subcomp->results[i]->atomnum_sub, result->score);
        for (j = 0; j < subcomp->results[i]->atomnum; j++){  /* SUBCOMPȰ㤦 */
          pairNO = result->map[j];
          atomNO = result->atomicNO[j];

          if (pairNO < 0) continue;
          getLabel(atomNO, buff);
          fprintf(fout, "            %-4d%3d:%-3s %3d:%-3s\n", count, pairNO + 1, buff, j + 1, buff);  /* SUBCOMPȰ㤦 */
          count++;
        }
        break;
      }
      fprintf(fout, "///\n");
    } else {
      fprintf(fout, "%s   %f\n", result->idstr, result->score);
    }
  }
}

void freeRESULTlist(RESULT **result_top){
  RESULT *p;
  RESULT *pNext;
  p = *result_top;

  while(p != NULL){
    pNext = p->next;

    free(p->map);
    free(p->atomicNO);
    free(p);

    p = pNext;
  }
}

